from django.db.models.signals import post_save
from django.dispatch import receiver

from cardroom.gamemaster import set_controller
from cardroom.models import CashGame


@receiver(post_save, sender=CashGame)
def set_cash_game_controller(
        sender,
        instance=None,
        created=False,
        **kwargs,
):
    set_controller(instance)



from django.contrib import admin
from django.urls import include, path, URLPattern, URLResolver
from rest_framework.routers import DefaultRouter

from cardroom.apps import CardroomConfig
from cardroom.utilities import get_admin, get_auth, get_felt
from cardroom.views import (
    CashGameFeltDataView,
    CashGameFeltView,
    CashGameViewSet,
    HandHistoryFeltDataView,
    HandHistoryFeltView,
    HandHistoryViewSet,
    PokerViewSet,
    TableViewSet,
)

app_name: str = CardroomConfig.name
router: DefaultRouter = DefaultRouter()
urlpatterns: list[URLPattern | URLResolver] = [
    path(
        'cash-games/<int:pk>/data/',
        CashGameFeltDataView.as_view(),
        name='cash-game-data',
    ),
    path(
        'hand-histories/<int:pk>/data/',
        HandHistoryFeltDataView.as_view(),
        name='hand-history-data',
    ),
]

router.register('poker', PokerViewSet)
router.register('tables', TableViewSet)
router.register('cash-games', CashGameViewSet)
router.register('hand-histories', HandHistoryViewSet)

if get_admin():
    urlpatterns.append(path('admin/', admin.site.urls))

if get_auth():
    urlpatterns.append(path('api-auth/', include('rest_framework.urls')))

if get_felt():
    urlpatterns.append(
        path(
            'cash-games/<int:pk>/felt/',
            CashGameFeltView.as_view(),
            name='cash-game-felt',
        ),
    )
    urlpatterns.append(
        path(
            'hand-histories/<int:pk>/felt/',
            HandHistoryFeltView.as_view(),
            name='hand-history-felt',
        ),
    )

urlpatterns.append(path('', include(router.urls)))



from abc import ABC, abstractmethod

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer

from cardroom.models import CashGame
from cardroom.gamemaster import handle, get_data
from cardroom.utilities import serialize


class ControllerConsumer(JsonWebsocketConsumer, ABC):
    @property
    def pk(self):
        return self.scope['url_route']['kwargs']['pk']

    @property
    def user(self):
        return self.scope['user']

    @property
    @abstractmethod
    def controller(self):
        pass

    def connect(self):
        super().connect()
        async_to_sync(self.channel_layer.group_add)(
            self.controller.group_name,
            self.channel_name,
        )
        self.data(
            {'type': 'data', 'data': (serialize(get_data(self.controller)),)},
        )

    def disconnect(self, code):
        async_to_sync(self.channel_layer.group_discard)(
            self.controller.group_name,
            self.channel_name,
        )
        super().disconnect(code)

    def receive_json(self, content, **kwargs):
        if self.user.is_authenticated:
            handle(self.controller, self.user, content)
        else:
            self.message(
                {
                    'type': 'message',
                    'user': '',
                    'message': 'you are unauthenticated',
                },
            )

    def data(self, event):
        self.send_json(
            (
                event
                | {
                    'data': [
                        data.get(
                            self.user.username,
                            data[''],
                        ) for data in event['data']
                    ],
                }
            ),
        )

    def message(self, event):
        if (
                (self.user.is_anonymous and not event['user'])
                or (
                    self.user.is_authenticated
                    and event['user'] == self.user.username
                )
        ):
            self.send_json(event)


class CashGameConsumer(ControllerConsumer):
    @property
    def controller(self):
        return CashGame.objects.get(pk=self.pk)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import fields
from functools import partial
from typing import ClassVar

from django.contrib.auth.models import AbstractUser
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext_lazy as _
from pokerkit import Automation, ValuesLike
import pokerkit

from cardroom.apps import CardroomConfig
from cardroom.felt import Data
from cardroom.gamemaster import broadcast, get_data
from cardroom.utilities import (
    get_divmod,
    get_felt,
    get_parse_value,
    get_root_routingconf,
    get_tzinfo,
)
import cardroom.controller as controller
import cardroom.table as table


class Variant(models.TextChoices):
    FIXED_LIMIT_TEXAS_HOLDEM = 'FT', _('Fixed-limit Texas hold \'em')
    NO_LIMIT_TEXAS_HOLDEM = 'NT', _('No-limit Texas hold \'em')
    NO_LIMIT_SHORT_DECK_HOLDEM = 'NS', _('No-limit short-deck hold \'em')
    POT_LIMIT_OMAHA_HOLDEM = 'PO', _('Pot-limit Omaha hold \'em')
    FIXED_LIMIT_OMAHA_HOLDEM_HIGH_LOW_SPLIT_EIGHT_OR_BETTER = (
        'FO/8',
        _('Fixed-limit Omaha hold \'em high/low-split eight-or-better'),
    )
    FIXED_LIMIT_SEVEN_CARD_STUD = 'F7S', _('Fixed-limit seven card stud')
    FIXED_LIMIT_SEVEN_CARD_STUD_HIGH_LOW_SPLIT_EIGHT_OR_BETTER = (
        'F7S/8',
        _('Fixed-limit seven card stud high/low-split eight-or-better'),
    )
    FIXED_LIMIT_RAZZ = 'FR', _('Fixed-limit razz')
    NO_LIMIT_DEUCE_TO_SEVEN_LOWBALL_SINGLE_DRAW = (
        'N2L1D',
        _('No-limit deuce-to-seven lowball single draw'),
    )
    FIXED_LIMIT_DEUCE_TO_SEVEN_LOWBALL_TRIPLE_DRAW = (
        'F2L3D',
        _('Fixed-limit deuce-to-seven lowball triple draw'),
    )
    FIXED_LIMIT_BADUGI = 'FB', _('Fixed-limit badugi')


class Poker(models.Model):
    automations: ClassVar[tuple[Automation]] = (Automation.CARD_BURNING,)
    variant = models.CharField(max_length=255, choices=Variant.choices)
    ante_trimming_status = models.BooleanField(default=False)
    raw_antes = models.JSONField()
    raw_blinds_or_straddles = models.JSONField(blank=True, null=True)
    bring_in = models.JSONField(blank=True, null=True)
    small_bet = models.JSONField(blank=True, null=True)
    big_bet = models.JSONField(blank=True, null=True)
    min_bet = models.JSONField(blank=True, null=True)

    def load(self) -> pokerkit.Poker:

        def clean(raw_values: ValuesLike) -> ValuesLike:
            if isinstance(raw_values, dict):
                raw_values = dict(
                    zip(map(int, raw_values.keys()), raw_values.values()),
                )

            return raw_values

        kwargs = {
            'automations': self.automations,
            'ante_trimming_status': self.ante_trimming_status,
            'raw_antes': clean(self.raw_antes),
            'raw_blinds_or_straddles': clean(self.raw_blinds_or_straddles),
            'bring_in': self.bring_in,
            'small_bet': self.small_bet,
            'big_bet': self.big_bet,
            'min_bet': self.min_bet,
            'divmod': get_divmod(),
        }

        for key, value in tuple(kwargs.items()):
            if value is None:
                kwargs.pop(key)

        return pokerkit.HandHistory.game_types[self.variant](**kwargs)

    class Meta:
        verbose_name_plural = 'poker'


class Table(models.Model):
    game = models.ForeignKey(Poker, models.PROTECT)
    seat_count = models.PositiveBigIntegerField()
    min_starting_stack = models.JSONField(blank=True, null=True)
    max_starting_stack = models.JSONField(blank=True, null=True)

    def load(self) -> table.Table:
        return table.Table(
            self.game.load(),
            self.seat_count,
            self.min_starting_stack,
            self.max_starting_stack,
        )


class Controller(models.Model):
    time_bank = models.FloatField()
    time_bank_increment = models.FloatField()
    state_construction_timeout = models.FloatField()
    state_destruction_timeout = models.FloatField()
    idle_timeout = models.FloatField()
    standing_pat_timeout = models.FloatField()
    betting_timeout = models.FloatField()
    hole_cards_showing_or_mucking_timeout = models.FloatField()

    @property
    def group_name(self) -> str:
        return f'{type(self).__name__}-{self.pk}'

    class Meta:
        abstract = True


class CashGame(Controller):
    table = models.ForeignKey(Table, models.PROTECT)

    def get_data(self, user: AbstractUser) -> Data:
        data = get_data(self)

        return data.get(user.username, data[''])

    def get_data_url(self) -> str:
        try:
            url = reverse(
                f'{CardroomConfig.name}:cash-game-data',
                kwargs={'pk': self.pk},
            )
        except NoReverseMatch:
            url = reverse('cash-game-data', kwargs={'pk': self.pk})

        return url

    def get_websocket_url(self) -> str:
        try:
            url = reverse(
                f'{CardroomConfig.name}:cash-game-websocket',
                get_root_routingconf(),
                kwargs={'pk': self.pk},
            )
        except NoReverseMatch:
            url = reverse(
                'cash-game-websocket',
                get_root_routingconf(),
                kwargs={'pk': self.pk},
            )

        return url

    if get_felt():
        def get_felt_url(self) -> str:
            try:
                url = reverse(
                    f'{CardroomConfig.name}:cash-game-felt',
                    kwargs={'pk': self.pk},
                )
            except NoReverseMatch:
                url = reverse('cash-game-felt', kwargs={'pk': self.pk})

            return url

        def get_absolute_url(self) -> str:
            return self.get_felt_url()

    def load(self) -> controller.CashGame:
        return controller.CashGame(
            self.time_bank,
            self.time_bank_increment,
            self.state_construction_timeout,
            self.state_destruction_timeout,
            self.idle_timeout,
            self.standing_pat_timeout,
            self.betting_timeout,
            self.hole_cards_showing_or_mucking_timeout,
            partial(broadcast, self),
            get_parse_value(),
            get_tzinfo(),
            self.table.load(),
        )


class HandHistory(models.Model):
    automations: ClassVar[tuple[Automation]] = (Automation.CARD_BURNING,)
    variant = models.CharField(max_length=255, choices=Variant.choices)
    ante_trimming_status = models.BooleanField(default=False)
    antes = models.JSONField()
    blinds_or_straddles = models.JSONField(blank=True, null=True)
    bring_in = models.JSONField(blank=True, null=True)
    small_bet = models.JSONField(blank=True, null=True)
    big_bet = models.JSONField(blank=True, null=True)
    min_bet = models.JSONField(blank=True, null=True)
    starting_stacks = models.JSONField()
    actions = models.JSONField()
    author = models.CharField(max_length=255, blank=True, null=True)
    event = models.CharField(max_length=255, blank=True, null=True)
    url = models.CharField(max_length=255, blank=True, null=True)
    address = models.CharField(max_length=255, blank=True, null=True)
    city = models.CharField(max_length=255, blank=True, null=True)
    region = models.CharField(max_length=255, blank=True, null=True)
    postal_code = models.CharField(max_length=255, blank=True, null=True)
    country = models.CharField(max_length=255, blank=True, null=True)
    time = models.TimeField(blank=True, null=True)
    time_zone = models.CharField(max_length=255, blank=True, null=True)
    day = models.PositiveBigIntegerField(blank=True, null=True)
    month = models.PositiveBigIntegerField(blank=True, null=True)
    year = models.PositiveBigIntegerField(blank=True, null=True)
    hand = models.PositiveBigIntegerField(blank=True, null=True)
    level = models.PositiveBigIntegerField(blank=True, null=True)
    seats = models.JSONField(blank=True, null=True)
    seat_count = models.PositiveBigIntegerField(blank=True, null=True)
    table = models.PositiveBigIntegerField(blank=True, null=True)
    players = models.JSONField(blank=True, null=True)
    finishing_stacks = models.JSONField(blank=True, null=True)
    currency = models.CharField(max_length=255, blank=True, null=True)
    time_limit = models.JSONField(blank=True, null=True)
    time_banks = models.JSONField(blank=True, null=True)

    @classmethod
    def get_field_names(cls) -> Iterator[str]:
        for field in fields(pokerkit.HandHistory):
            try:
                cls._meta.get_field(field.name)
            except FieldDoesNotExist:
                pass
            else:
                yield field.name

    @property
    def data(self) -> Iterator[Data]:
        return Data.from_hand_history(self.load())

    def get_data_url(self) -> str:
        try:
            url = reverse(
                f'{CardroomConfig.name}:hand-history-data',
                kwargs={'pk': self.pk},
            )
        except NoReverseMatch:
            url = reverse('hand-history-data', kwargs={'pk': self.pk})

        return url

    if get_felt():
        def get_felt_url(self) -> str:
            try:
                url = reverse(
                    f'{CardroomConfig.name}:hand-history-felt',
                    kwargs={'pk': self.pk},
                )
            except NoReverseMatch:
                url = reverse('hand-history-felt', kwargs={'pk': self.pk})

            return url

        def get_absolute_url(self) -> str:
            return self.get_felt_url()

    @classmethod
    def dump(cls, hh: pokerkit.HandHistory) -> HandHistory:
        kwargs = {}

        for name in cls.get_field_names():
            kwargs[name] = getattr(hh, name)

        return cls(**kwargs)

    def load(self) -> pokerkit.HandHistory:
        kwargs = {}

        for name in type(self).get_field_names():
            kwargs[name] = getattr(self, name)

        return pokerkit.HandHistory(**kwargs, automations=self.automations)

    class Meta:
        verbose_name_plural = 'hand histories'



from collections.abc import Callable, Iterable, Mapping
from dataclasses import asdict, is_dataclass
from math import inf, floor
from typing import Any, cast
from zoneinfo import ZoneInfo
import math

from django.conf import settings
from django.utils.module_loading import import_string
import pokerkit

from cardroom.felt import Configuration

DEFAULT_DIVMOD: str = 'cardroom.utilities.divmod'
DEFAULT_PARSE_VALUE: str = 'cardroom.utilities.parse_value'
DEFAULT_DECIMAL_PLACES: int | float = inf
DEFAULT_ADMIN: bool = False
DEFAULT_AUTH: bool = False
DEFAULT_FELT: bool = False
DEFAULT_CONFIGURATION: Configuration = Configuration()
DEFAULT_ROOT_ROUTINGCONF: str = 'cardroom.routing'
_divmod = divmod


def divmod(dividend: int, divisor: int) -> tuple[int, int]:
    quotient: int | float
    remainder: int | float

    match get_decimal_places():
        case 0:
            quotient, remainder = _divmod(int(dividend), int(divisor))
        case math.inf:
            quotient, remainder = dividend / divisor, 0
        case decimal_places:
            assert isinstance(decimal_places, int)

            quotient = (
                floor(dividend / divisor * 10 ** decimal_places)
                / (10 ** decimal_places)
            )
            remainder = dividend - quotient * divisor

    return cast(tuple[int, int], (quotient, remainder))


def parse_value(raw_value: str) -> int:
    value = pokerkit.parse_value(raw_value)

    match get_decimal_places():
        case 0:
            value = int(value)
        case math.inf:
            pass
        case decimal_places:
            assert isinstance(decimal_places, int)

            value = round(value, decimal_places)

    return value


def get_tzinfo() -> ZoneInfo:
    return ZoneInfo(getattr(settings, 'TIME_ZONE', 'UTC'))


def get_divmod() -> Callable[[int, int], tuple[int, int]]:
    return import_string(getattr(settings, 'CARDROOM_DIVMOD', DEFAULT_DIVMOD))


def get_parse_value() -> Callable[[str], int]:
    return import_string(
        getattr(settings, 'CARDROOM_PARSE_VALUE', DEFAULT_PARSE_VALUE),
    )


def get_decimal_places() -> int | float:
    return getattr(settings, 'CARDROOM_DECIMAL_PLACES', DEFAULT_DECIMAL_PLACES)


def get_admin() -> bool:
    return getattr(settings, 'CARDROOM_ADMIN', DEFAULT_ADMIN)


def get_auth() -> bool:
    return getattr(settings, 'CARDROOM_AUTH', DEFAULT_AUTH)


def get_felt() -> bool:
    return getattr(settings, 'CARDROOM_FELT', DEFAULT_FELT)


def get_configuration() -> Configuration:
    return getattr(settings, 'CARDROOM_CONFIGURATION', DEFAULT_CONFIGURATION)


def get_root_routingconf() -> Any:
    return getattr(settings, 'ROOT_ROUTINGCONF', DEFAULT_ROOT_ROUTINGCONF)


def serialize(obj: Any) -> Any:
    if obj is None or isinstance(obj, bytes | str | int | float | bool):
        return obj
    elif is_dataclass(obj):
        return serialize(asdict(obj))
    elif isinstance(obj, Mapping):
        return dict(zip(serialize(obj.keys()), serialize(obj.values())))
    elif isinstance(obj, Iterable):
        return list(map(serialize, obj))
    else:
        raise AssertionError



#!/usr/bin/env python3

from setuptools import find_packages, setup

setup(
    name='cardroom',
    version='0.0.1.dev6',
    description='A Django application for poker tournament and table management',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/uoftcprg/cardroom',
    author='University of Toronto Computer Poker Research Group',
    author_email='uoftcprg@outlook.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Topic :: Education',
        'Topic :: Games/Entertainment',
        'Topic :: Games/Entertainment :: Board Games',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    keywords=[
        'channels',
        'django',
        'game',
        'game-development',
        'holdem-poker',
        'imperfect-information-game',
        'poker',
        'poker-engine',
        'poker-game',
        'poker-library',
        'poker-strategies',
        'python',
        'texas-holdem',
    ],
    project_urls={
        'Documentation': 'https://cardroom.readthedocs.io/en/latest/',
        'Source': 'https://github.com/uoftcprg/cardroom',
        'Tracker': 'https://github.com/uoftcprg/cardroom/issues',
    },
    packages=find_packages(),
    install_requires=[
        'channels[daphne]>=4.0.0,<5',
        'Django>=4.2.9,<5',
        'djangorestframework>=3.14.0,<4',
        'pokerkit~=0.4.8',
    ],
    python_requires='>=3.11',
    package_data={'cardroom': ['py.typed', 'static/**/*', 'templates/**/*']},
)



from django.apps import AppConfig
from django.db import OperationalError


class CardroomConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cardroom'

    def ready(self) -> None:
        from cardroom.models import CashGame
        from cardroom.gamemaster import set_controller

        __import__('cardroom.signals')

        try:
            cash_games = list(CashGame.objects.all())
        except OperationalError:
            cash_games = []

        for cash_game in cash_games:
            set_controller(cash_game)
